@@ -1,5 +1,5 @@
 # - perl *-
-# $Id: cssh.pl 273 2009-06-02 15:16:42Z duncan_ferguson $
+# $Id$
 #
 # Script:
 #   $RCSfile$
@@ -50,12 +50,7 @@
 #     http://www.opensource.org/licenses/gpl-license.php
 #
 ############################################################################
-my $VERSION = '$Revision: 273 $ ($Date: 2009-06-02 16:16:42 +0100 (Tue, 02 Jun 2009) $)';
-
-# Now tidy it up, but in such as way cvs doesn't kill the tidy up stuff
-$VERSION =~ s/\$Revision: //;
-$VERSION =~ s/\$Date: //;
-$VERSION =~ s/ \$//g;
+my $VERSION = '3.28';
 
 ### all use statements ###
 use strict;
@@ -81,6 +76,8 @@ use X11::Keysyms '%keysymtocode', 'MISCELLANY', 'XKB_KEYS', '3270', 'LATIN1',
 use File::Basename;
 use Net::hostent;
 use Carp;
+use Sys::Hostname;
+use English;
 
 ### all global variables ###
 my $scriptname = $0;
@@ -266,6 +263,12 @@ sub load_config_defaults() {
 
     $config{command}             = q{};
     $config{max_host_menu_items} = 30;
+
+    $config{max_addhost_menu_cluster_items} = 6;
+    $config{menu_send_autotearoff}          = 0;
+    $config{menu_host_autotearoff}          = 0;
+
+    $config{send_menu_xml_file} = $ENV{HOME} . '/.csshrc_send_menu';
 }
 
 # load in config file settings
@@ -275,13 +278,20 @@ sub parse_config_file($) {
     return if ( !-e $config_file || !-r $config_file );
 
     open( CFG, $config_file ) or die("Couldnt open $config_file: $!");
-    while (<CFG>) {
-        next if ( /^\s*$/ || /^#/ );    # ignore blank lines & commented lines
-        s/#.*//;                        # remove comments from remaining lines
-        s/\s*$//;                       # remove trailing whitespace
-        chomp();
+    my $l;
+    while ( defined( $l = <CFG> ) ) {
+        next
+            if ( $l =~ /^\s*$/ || $l =~ /^#/ )
+            ;    # ignore blank lines & commented lines
+        $l =~ s/#.*//;     # remove comments from remaining lines
+        $l =~ s/\s*$//;    # remove trailing whitespace
+        chomp $l;
+        if ( $l =~ s/\\\s*$// ) {
+            $l .= <CFG>;
+            redo unless eof(CFG);
+        }
 
-        next unless m/\s*(\S+)\s*=\s*(.*)\s*/;
+        next unless $l =~ m/\s*(\S+)\s*=\s*(.*)\s*/;
         my ( $key, $value ) = ( $1, $2 );
         $config{$key} = $value;
         logmsg( 3, "$key=$value" );
@@ -297,7 +307,7 @@ sub find_binary($) {
 
     logmsg( 2, "Looking for $binary" );
     my $path;
-    if ( !-x $binary ) {
+    if ( !-x $binary || substr($binary, 0, 1) ne '/' ) {
 
        # search the users $PATH and then a few other places to find the binary
        # just in case $PATH isnt set up right
@@ -316,7 +326,7 @@ sub find_binary($) {
         {
             logmsg( 3, "Looking in $_" );
 
-            if ( -x $_ . '/' . $binary ) {
+            if ( -f $_ . '/' . $binary && -x $_ . '/' . $binary ) {
                 $path = $_ . '/' . $binary;
                 logmsg( 2, "Found at $path" );
                 last;
@@ -430,7 +440,9 @@ sub check_ssh_hostnames {
     if ( -r $ssh_config && open( SSHCFG, "<", $ssh_config ) ) {
         while (<SSHCFG>) {
             next unless (m/^\s*host\s+([\w\.-]+)/i);
-            $ssh_hostnames{$1} = 1;
+
+            # account for multiple declarations of hosts
+            $ssh_hostnames{$_} = 1 foreach ( split( /\s+/, $1 ) );
         }
         close(SSHCFG);
     }
@@ -604,11 +616,17 @@ sub get_clusters() {
     if ( -f $cluster_file ) {
         logmsg( 2, "Loading clusters in from $cluster_file" );
         open( CLUSTERS, $cluster_file ) || die("Couldnt read $cluster_file");
-        while (<CLUSTERS>) {
+        my $l;
+        while ( defined( $l = <CLUSTERS> ) ) {
             next
-                if ( /^\s*$/ || /^#/ ); # ignore blank lines & commented lines
-            chomp();
-            my @line = split(/\s/);
+                if ( $l =~ /^\s*$/ || $l =~ /^#/ )
+                ;    # ignore blank lines & commented lines
+            chomp $l;
+            if ( $l =~ s/\\\s*$// ) {
+                $l .= <CLUSTER>;
+                redo unless eof(CLUSTERS);
+            }
+            my @line = split( /\s/, $l );
 
         #s/^([\w-]+)\s*//;               # remote first word and stick into $1
 
@@ -618,7 +636,7 @@ sub get_clusters() {
                 join( " ", @line[ 1 .. $#line ] )
             );
             $clusters{ $line[0] } = join( " ", @line[ 1 .. $#line ] )
-                ;                       # Now bung in rest of line
+                ;    # Now bung in rest of line
         }
         close(CLUSTERS);
     }
@@ -660,11 +678,16 @@ sub get_clusters() {
                 logmsg( 2, "Loading clusters in from '$file'" );
 
                 open( CLUSTERS, $file ) || die("Couldnt read '$file': $!\n");
-                while (<CLUSTERS>) {
-                    next if ( /^\s*$/ || /^#/ );
-                    chomp;
+                my $l;
+                while ( defined( $l = <CLUSTERS> ) ) {
+                    next if ( $l =~ /^\s*$/ || $l =~ /^#/ );
+                    chomp $l;
+                    if ( $l =~ s/\\\s*$// ) {
+                        $l .= <CLUSTER>;
+                        redo unless eof(CLUSTERS);
+                    }
 
-                    my @line = split(/\s/);
+                    my @line = split( /\s/, $l );
                     logmsg(
                         3,
                         "cluster $line[0] = ",
@@ -682,30 +705,36 @@ sub get_clusters() {
 }
 
 sub resolve_names(@) {
-    logmsg( 2, "Resolving cluster names: started" );
+    logmsg( 2, 'Resolving cluster names: started' );
     my @servers = @_;
 
     foreach (@servers) {
-        logmsg( 3, "Found server $_" );
+        my $dirty    = $_;
+        my $username = q{};
+        logmsg( 3, 'Checking tag ',$_ );
 
-        if ( $clusters{$_} ) {
-            push( @servers, split( / /, $clusters{$_} ) );
-            $_ = "";
+        if ( $dirty =~ s/^(.*)@// ) {
+            $username = $1;
+        }
+        if ( $clusters{$dirty} ) {
+            logmsg( 3, '... it is a cluster' );
+            foreach my $node ( split( / /, $clusters{$dirty} ) ) {
+                if($username) {
+                    $node =~ s/^(.*)@//;
+                    $node = $username . '@' . $node;
+                }
+                push( @servers, $node );
+            }
+            $_ = q{};
         }
     }
 
-    my @cleanarray;
-
     # now clean the array up
-    foreach (@servers) {
-        push( @cleanarray, $_ ) if ( $_ !~ /^$/ );
-    }
+    @servers = grep { $_ !~ m/^$/} @servers;
 
-    foreach (@cleanarray) {
-        logmsg( 3, "leaving with $_" );
-    }
-    logmsg( 2, "Resolving cluster names: completed" );
-    return (@cleanarray);
+    logmsg( 3, 'leaving with ', $_ ) foreach (@servers);
+    logmsg( 2, 'Resolving cluster names: completed' );
+    return (@servers);
 }
 
 sub change_main_window_title() {
@@ -769,9 +798,26 @@ sub send_text($@) {
     my $svr = shift;
     my $text = join( "", @_ );
 
-    #logmsg( 2, "Sending to $svr text:$text:" );
-
     logmsg( 2, "servers{$svr}{wid}=$servers{$svr}{wid}" );
+    logmsg( 3, "Sending to '$svr' text:$text:" );
+
+    # command macro substitution
+
+    # $svr contains a trailing space here, so ensure its stripped off
+    {
+        my $servername = $svr;
+        $servername =~ s/\s+//;
+        $text       =~ s/%s/$servername/xsm;
+    }
+    $text =~ s/%h/hostname()/xsme;
+
+    # use connection username, else default to current username
+    {
+        my $username = $servers{$svr}{username};
+        $username ||= getpwuid($UID);
+        $text =~ s/%u/$username/xsm;
+    }
+    $text =~ s/%n/\n/xsm;
 
     foreach my $char ( split( //, $text ) ) {
         next if ( !defined($char) );
@@ -813,9 +859,11 @@ sub send_text($@) {
     $xdisplay->flush();
 }
 
-sub send_clientname() {
+sub send_text_to_all_servers {
+    my $text = join( '', @_ );
+
     foreach my $svr ( keys(%servers) ) {
-        send_text( $svr, $servers{$svr}{givenname} )
+        send_text( $svr, $text )
             if ( $servers{$svr}{active} == 1 );
     }
 }
@@ -984,12 +1032,11 @@ sub check_host($) {
         else {
             logmsg( 1,
                 "Failed to check host (falling back to gethostbyname): $!" );
-            return gethostbyname($host)->name();
         }
     }
-    else {
-        return gethostbyname($host)->name();
-    }
+
+    my $gethost_obj = gethostbyname($host);
+    return defined($gethost_obj) ? $gethost_obj->name() : $host;
 }
 
 sub open_client_windows(@) {
@@ -1059,7 +1106,7 @@ sub open_client_windows(@) {
           # affecting the main program
             $servers{$server}{realname} .= "==" if ( !$realname );
             my $exec
-                = "$config{terminal} $color $config{terminal_args} $config{terminal_allow_send_events} $config{terminal_title_opt} '$config{title}: $servers{$server}{connect_string}' -font $config{terminal_font} -e \"$^X\" \"-e\" '$helper_script' '$servers{$server}{pipenm}' '$servers{$server}{realname}' '$servers{$server}{username}' '$servers{$server}{port}'";
+                = "$config{terminal} $color $config{terminal_args} $config{terminal_allow_send_events} $config{terminal_title_opt} '$config{title}: $servers{$server}{connect_string}' -font $config{terminal_font} -e \"$^X\" \"-e\" '$helper_script' '$servers{$server}{pipenm}' '$servers{$server}{givenname}' '$servers{$server}{username}' '$servers{$server}{port}'";
             logmsg( 2, "Terminal exec line:\n$exec\n" );
             exec($exec) == 0 or warn("Failed: $!");
         }
@@ -1159,6 +1206,15 @@ sub show_console() {
 
     # Sleep for a moment to give WM time to bring console back
     select( undef, undef, undef, 0.5 );
+
+    if ( $config{menu_send_autotearoff} ) {
+        $menus{send}->menu->tearOffMenu()->raise;
+    }
+
+    if ( $config{menu_host_autotearoff} ) {
+        $menus{hosts}->menu->tearOffMenu()->raise;
+    }
+
     $windows{main_window}->deiconify;
     $windows{main_window}->raise;
     $windows{main_window}->focus( -force );
@@ -1474,10 +1530,17 @@ sub add_host_by_name() {
         return;
     }
 
-    logmsg( 2, "host=$menus{host_entry}" );
+    if ( $menus{host_entry} ) {
+        logmsg( 2, "host=", $menus{host_entry} );
+        open_client_windows(
+            resolve_names( split( /\s+/, $menus{host_entry} ) ) );
+    }
 
-    open_client_windows(
-        resolve_names( split( /\s+/, $menus{host_entry} ) ) );
+    if ( $menus{listbox}->curselection() ) {
+        my @hosts = $menus{listbox}->get( $menus{listbox}->curselection() );
+        logmsg( 2, "host=", join( ' ', @hosts ) );
+        open_client_windows( resolve_names(@hosts) );
+    }
 
     build_hosts_menu();
     $menus{host_entry} = "";
@@ -1671,7 +1734,9 @@ sub create_windows() {
             my $paste_text = '';
 
             # SelectionGet is fatal if no selection is given
-            Tk::catch { $paste_text = $windows{main_window}->SelectionGet };
+            Tk::catch {
+                $paste_text = $windows{main_window}->SelectionGet;
+            };
 
             if ( !length($paste_text) ) {
                 warn("Got empty paste event\n");
@@ -1724,6 +1789,27 @@ sub create_windows() {
         -default_button => 'Add',
     );
 
+    if ( $config{max_addhost_menu_cluster_items}
+        && scalar keys %clusters )
+    {
+        if ( scalar keys %clusters < $config{max_addhost_menu_cluster_items} )
+        {
+            $menus{listbox} = $windows{addhost}->Listbox(
+                -selectmode => 'extended',
+                -height     => scalar keys %clusters,
+            )->pack();
+        }
+        else {
+            $menus{listbox} = $windows{addhost}->Scrolled(
+                'Listbox',
+                -scrollbars => 'e',
+                -selectmode => 'extended',
+                -height     => $config{max_addhost_menu_cluster_items},
+            )->pack();
+        }
+        $menus{listbox}->insert( 'end', sort keys %clusters );
+    }
+
     $windows{host_entry} = $windows{addhost}->add(
         'LabEntry',
         -textvariable => \$menus{host_entry},
@@ -1773,41 +1859,41 @@ sub capture_map_events() {
         }
     );
 
-#    $windows{main_window}->bind(
-#        '<Unmap>' => sub {
-#            logmsg( 3, "Entering UNMAP" );
-#
-#            my $state = $windows{main_window}->state();
-#            logmsg( 3,
-#                "state=$state previous=$config{internal_previous_state}" );
-#
-#            if ( $config{internal_previous_state} eq $state ) {
-#                logmsg( 3, "repeating the same" );
-#            }
-#
-#            if ( $config{internal_previous_state} eq "mid-change" ) {
-#                logmsg( 3, "dropping out as mid-change" );
-#                return;
-#            }
-#
-#            if ( $config{internal_previous_state} eq "normal" ) {
-#                logmsg( 3, "withdrawing all windows" );
-#                foreach my $server ( reverse( keys(%servers) ) ) {
-#                    $xdisplay->req( 'UnmapWindow', $servers{$server}{wid} );
-#                    if ( $config{unmap_on_redraw} =~ /yes/i ) {
-#                        $xdisplay->req( 'UnmapWindow',
-#                            $servers{$server}{wid} );
-#                    }
-#                }
-#                $xdisplay->flush();
-#            }
-#
-#            if ( $config{internal_previous_state} ne $state ) {
-#                logmsg( 3, "resetting prev_state" );
-#                $config{internal_previous_state} = $state;
-#            }
-#        }
-#    );
+ #    $windows{main_window}->bind(
+ #        '<Unmap>' => sub {
+ #            logmsg( 3, "Entering UNMAP" );
+ #
+ #            my $state = $windows{main_window}->state();
+ #            logmsg( 3,
+ #                "state=$state previous=$config{internal_previous_state}" );
+ #
+ #            if ( $config{internal_previous_state} eq $state ) {
+ #                logmsg( 3, "repeating the same" );
+ #            }
+ #
+ #            if ( $config{internal_previous_state} eq "mid-change" ) {
+ #                logmsg( 3, "dropping out as mid-change" );
+ #                return;
+ #            }
+ #
+ #            if ( $config{internal_previous_state} eq "normal" ) {
+ #                logmsg( 3, "withdrawing all windows" );
+ #                foreach my $server ( reverse( keys(%servers) ) ) {
+ #                    $xdisplay->req( 'UnmapWindow', $servers{$server}{wid} );
+ #                    if ( $config{unmap_on_redraw} =~ /yes/i ) {
+ #                        $xdisplay->req( 'UnmapWindow',
+ #                            $servers{$server}{wid} );
+ #                    }
+ #                }
+ #                $xdisplay->flush();
+ #            }
+ #
+ #            if ( $config{internal_previous_state} ne $state ) {
+ #                logmsg( 3, "resetting prev_state" );
+ #                $config{internal_previous_state} = $state;
+ #            }
+ #        }
+ #    );
 }
 
 # for all key event, event hotkeys so there is only 1 key binding
@@ -1848,11 +1934,14 @@ sub key_event {
             if ( $combo =~ /^$key$/ ) {
                 if ( $event eq "KeyRelease" ) {
                     logmsg( 2, "Received hotkey: $hotkey" );
-                    send_clientname()     if ( $hotkey eq "key_clientname" );
-                    add_host_by_name()    if ( $hotkey eq "key_addhost" );
-                    retile_hosts("force") if ( $hotkey eq "key_retilehosts" );
-                    show_history()        if ( $hotkey eq "key_history" );
-                    exit_prog()           if ( $hotkey eq "key_quit" );
+                    send_text_to_all_servers('%s')
+                        if ( $hotkey eq "key_clientname" );
+                    add_host_by_name()
+                        if ( $hotkey eq "key_addhost" );
+                    retile_hosts("force")
+                        if ( $hotkey eq "key_retilehosts" );
+                    show_history() if ( $hotkey eq "key_history" );
+                    exit_prog()    if ( $hotkey eq "key_quit" );
                 }
                 return;
             }
@@ -1860,7 +1949,8 @@ sub key_event {
     }
 
     # look for a <Control>-d and no hosts, so quit
-    exit_prog() if ( $state =~ /Control/ && $keysym eq "d" and !%servers );
+    exit_prog()
+        if ( $state =~ /Control/ && $keysym eq "d" and !%servers );
 
     update_display_text( $keycodetosym{$keysymdec} )
         if ( $event eq "KeyPress" && $keycodetosym{$keysymdec} );
@@ -1944,17 +2034,12 @@ sub create_menubar() {
     );
 
     $menus{send} = $menus{bar}->cascade(
-        -label     => 'Send',
-        -menuitems => [
-            [   "command",
-                "Hostname",
-                -command     => \&send_clientname,
-                -accelerator => $config{key_clientname},
-            ],
-        ],
+        -label   => 'Send',
         -tearoff => 1,
     );
 
+    populate_send_menu();
+
     $menus{help} = $menus{bar}->cascade(
         -label     => 'Help',
         -menuitems => [
@@ -1974,11 +2059,84 @@ sub create_menubar() {
     logmsg( 2, "create_menubar: completed" );
 }
 
+sub populate_send_menu_entries_from_xml {
+    my ( $menu, $menu_xml ) = @_;
+
+    foreach my $menu_ref ( @{ $menu_xml->{menu} } ) {
+        if ( $menu_ref->{menu} ) {
+            $menus{ $menu_ref->{title} }
+                = $menu->cascade( -label => $menu_ref->{title}, );
+            populate_send_menu_entries_from_xml( $menus{ $menu_ref->{title} },
+                $menu_ref, );
+            if ( $menu_ref->{detach} && $menu_ref->{detach} =~ m/y/i ) {
+                $menus{ $menu_ref->{title} }->menu->tearOffMenu()->raise;
+            }
+        }
+        else {
+            my $command     = undef;
+            my $accelerator = undef;
+            if ( $menu_ref->{command} ) {
+                $command = sub {
+                    send_text_to_all_servers( $menu_ref->{command}[0] );
+                };
+            }
+            if ( $menu_ref->{accelerator} ) {
+                $accelerator = $menu_ref->{accelerator};
+            }
+            $menu->command(
+                -label       => $menu_ref->{title},
+                -command     => $command,
+                -accelerator => $accelerator,
+            );
+        }
+    }
+
+    return;
+}
+
+sub populate_send_menu {
+
+    #    my @menu_items = ();
+    if ( !-r $config{send_menu_xml_file} ) {
+        logmsg( 2, 'Using default send menu' );
+
+        $menus{send}->command(
+            -label       => 'Hostname',
+            -command     => [ \&send_text_to_all_servers, '%s' ],
+            -accelerator => $config{key_clientname},
+        );
+    }
+    else {
+        logmsg(
+            2,
+            'Using xml send menu definition from ',
+            $config{send_menu_xml_file}
+        );
+
+        eval { require XML::Simple; };
+        die 'Cannot load XML::Simple - has it been installed?  ', $@ if ($@);
+
+        my $xml = XML::Simple->new( ForceArray => 1, );
+        my $menu_xml = $xml->XMLin( $config{send_menu_xml_file} );
+
+        logmsg( 3, 'xml send menu: ', $/, $xml->XMLout($menu_xml) );
+
+        if ( $menu_xml->{detach} && $menu_xml->{detach} =~ m/y/i ) {
+            $menus{send}->menu->tearOffMenu()->raise;
+        }
+
+        populate_send_menu_entries_from_xml( $menus{send}, $menu_xml );
+    }
+
+    return;
+}
+
 ### main ###
 
 # Note: getopts returned "" if it finds any options it doesnt recognise
 # so use this to print out basic help
-pod2usage( -verbose => 1 ) if ( !GetOptions( \%options, @options_spec ) );
+pod2usage( -verbose => 1 )
+    if ( !GetOptions( \%options, @options_spec ) );
 pod2usage( -verbose => 1 ) if ( $options{'?'} || $options{help} );
 pod2usage( -verbose => 2 ) if ( $options{H}   || $options{man} );
 
@@ -2311,6 +2469,9 @@ on standard port (e.g not listening on port 22) and ssh_config cannot be used.
 Open a series of xterms defined by <tag> within either /etc/clusters or
 F<$HOME/.csshrc> (see L<"FILES">).
 
+Note: specifying a username on a cluster tag will override any usernames 
+defined in the cluster
+
 =back
 
 =head1 KEY SHORTCUTS
@@ -2382,7 +2543,7 @@ S<$ crsh server1 server2>
 
 =over
 
-=item /etc/clusters
+=item F</etc/clusters>
 
 This file contains a list of tags to server names mappings.  When any name
 is used on the command line it is checked to see if it is a tag.
@@ -2492,10 +2653,22 @@ See below notes on shortcuts.
 
 Default key sequence to retile host windows.  See below notes on shortcuts.
 
+=item max_addhost_menu_cluster_items = 6
+
+Maximum number of entries in the 'Add Host' menu cluster list before 
+scrollbars are used
+
 =item max_host_menu_items = 30
 
 Maximum number of hosts to put into the host menu before starting a new column
 
+=item menu_host_autotearoff = 0
+
+=item menu_send_autotearoff = 0
+
+When set to non-0 will automatically tear-off the host or send menu at 
+program start
+
 =item mouse_paste = Button-2 (middle mouse button)
 
 Default key sequence to paste text into the console window using the mouse.  
@@ -2623,6 +2796,67 @@ B<NOTE:> The key shortcut modifiers must be in the form "Control", "Alt", or
 "Shift", i.e. with the first letter capitalised and the rest lower case.  Keys
 may also be disabled individually by setting to the word "null".
 
+=item F<$HOME/.csshrc_send_menu>
+
+This (optional) file contains items to populate the send menu.  The
+default entry could be written as:
+
+  <send_menu>
+    <menu title="Hostname">
+        <command>%s</command>
+        <accelerator>ALT-n</accelerator>
+    </menu>
+  </send_menu>
+
+Submenus can also be specified as follows:
+
+  <send_menu>
+    <menu title="Default Entries">
+      <detach>yes</detach>
+      <menu title="Hostname">
+          <command>%s</command>
+          <accelerator>ALT-n</accelerator>
+      </menu>
+    </menu>
+  </send_menu>
+
+B<Caveats:> 
+
+=over 4
+
+=item There is currently no strict format checking of this file.
+
+=item The format of the file may change in the future
+
+=item If the file exists the default entry (Hostname) is not added
+
+=back
+
+The following replacement macros are available:
+
+=over 4
+
+=item %s 
+
+Hostname part of the specific connection string to each client, minus any 
+username or port
+
+=item %u
+
+Username part of the connection string to each client
+
+=item %h 
+
+Hostname of server where cssh is being run from
+
+=item %n
+
+<RETURN> code
+
+=back
+
+B<NOTE:> requires L<XML::Simple> to be installed
+
 =back
 
 =head1 KNOWN BUGS
